/*
Asynkronista JS:ää callbackillä ja promisella. Molemmissa on käytetty
setTimeout() -metodia joka aiheuttaa sen sisällä olevan koodin suoritukseen viiveen.
Tavallisesti viiveen aiheuttaja on esim. jonkin resurssin lataaminen web-palvelimelta.
Viiveen mittaista aikaa ei jäädä odottelemaan, vaan muun koodin suoritus jatkuu.
Promise käsitellään kahdella eri tavalla: then() -funktiolla ja async-await -rakenteella.
*/

/*
Callback-funktio on perinteinen tapa toteuttaa asynkroninen JS-koodi
Callback on edelleen paras tapa yksinkertaisissa tapauksissa jolloin
promisen avulla tehty koodi on pitempi ja monimutkaisempi.

Callback on funktio joka annetaan argumenttina toiselle funktiolle ja
tarkoituksena on että isäntäfunktion suoritus "laukaisee" callbackin
myöhemmässä vaiheessa.
*/
console.log('Turhapuroilla laitetaan ruokaa, mutta maito on lopussa!');

function maitoKauppa(callback) {
  console.log(
    'Johtaja Uuno Turhapuro lupaa hakea kaupasta maitoa callbackillä!'
  );
  callback();
}

maitoKauppa(function () {
  setTimeout(() => {
    console.log('Uuno suorittaa callbackia');
    try {
      if (Math.floor(Math.random() * 2) === 0) {
        // se on fifty-fifty miten käy
        console.log('Callbackin tulos oli maitoa.');
      } else {
        if (Math.floor(Math.random() * 2) === 0) {
          // se on taas fifty-fifty miten käy
          console.log('Callbackin tulos oli olutta.');
        } else {
          //'abc'.repeat(-1); // tällä voi kaataa callbackin suorituksen RangeError-virheeseen
          throw new Error('Uuno katosi täydellisesti ja menetti muistinsa.');
        }
      }
    } catch (e) {
      console.log('Callback tuotti virheen: ' + e); // ei saada maitoa eikä olutta kun homma kaatuu
    }
  }, 1000); // sekunnin mittainen suoritusaika
});

// Muun koodin suoritus voi jatkua vaikka callback onkin suorituksessa
// koodi toimii siis asynkronisesti
console.log(
  'Uunon vaimo jatkaa ruoanlaittoa vaikka Uuno lähti callbackillä maidonhakureissulle...'
);

/*
Promise on ES6:n uusi menetelmä asynkroniseen ohjelmointiin.
Promise kehitettiin korvaamaan callbackit sellaisissa tilanteissa joissa
callbackien käyttö aiheuttaa hankaluuksia. Monimutkaisessa koodissa promise
voi olla yksinkertaisempi kuin callbackit.

Promise on olio joka säilyttää asynkronisen operaation tulosta. Se edustaa arvoa
jota meillä ei vielä ole, mutta jos hyvin käy, arvo saadaan tulevaisuudessa.
Promisella on kolme tilaa: Suorituksessa(Pending), Toteutettu(Resolved), Hylätty(Rejected)
*/
const maitoLupaus = new Promise((resolve, reject) => {
  console.log('Härski Hartikainen lupaa hakea maitoa kaupasta promisella!');
  setTimeout(() => {
    console.log('Härski suorittaa promisea.');
    try {
      if (Math.floor(Math.random() * 2) === 0) {
        // se on fifty-fifty miten käy
        resolve('maitoa.');
      } else {
        if (Math.floor(Math.random() * 2) === 0) {
          // se on taas fifty-fifty miten käy
          resolve('olutta.');
        } else {
          //'abc'.repeat(-1); // tällä voi kaataa callbackin suorituksen RangeError-virheeseen
          throw new Error('Härski katosi täydellisesti ja menetti muistinsa.');
        }
      }
    } catch (e) {
      reject(e); // ei saada maitoa eikä olutta kun koko homma kaatuu
    }
  }, 1000); // sekunnin mittainen suoritusaika
});

// Promise käsitellään then-metodilla
maitoLupaus
  .then((tulos) => {
    console.log('Promise toteutui. Tulos oli ' + tulos);
  })
  .catch((e) => console.log('Promise ei toteutunut. ' + e));

// Muun koodin suoritus voi jatkua vaikka Promise onkin suorituksessa (Pending)
// koodi toimii siis asynkronisesti
console.log(
  'Uunon vaimo jatkaa ruoanlaittoa vaikka Härski lähti promisella maidonhakureissulle...'
);

/*Async-await on uusi tapa käsitellä promise. Sitä voidaan siis käyttää ainoastaan
valmiina saadun promisen käsittelyyn. Ohjelmakirjastojen asynkroniset metodit
palauttavat yleensä promiseja.
*/
async function maitoKauppaAsync() {
  console.log('Promise käsitellään async-awaitilla');
  try {
    const tulos = await maitoLupaus;
    console.log('Promise toteutui async-awaitilla. Tulos oli ' + tulos);
  } catch (e) {
    console.log('Promise ei toteutunut async-awaitilla. ' + e);
  }
}

maitoKauppaAsync();
